<!doctype html>
<html>
    <head>
        <meta charset="utf-8" />
        <title>{{ title }}</title>
    </head>
    <body>
        <div id="app"></div>
        <script src="static/bundle.js" type="text/javascript"></script>
    </body>
</html>
